{{//////////////////////////////////////////////////////////////////////

W5100 Ethernet Driver
Author: Shane Avery
LAST MODIFIED: 8.18.08
VERSION 1.1

Spin Functions:
        start : okay
                Starts the ASM in a new COG. 
                returns: okay - determines if the cog started ok or not

        stop
                Stops driver - frees a cog
                args:    none
                returns: none

        init (gway, subnet, ip, mac)
                Inits some registers.
                args:    gway - 4 byte array which is the gateway IP
                         subnet - 4 byte array which is the subnet
                         ip - 4 byte array which is this device's IP
                         mac - 6 byte array that is the MAC address
                returns: none                            
                

        open (type, source_port, dest_port, dest_ip)
                Opens Socket0.
                args:    type - 0 for UDP or 1 for TCP
                         source_port - source port
                         dest_port - desination port
                         dest_ip - destination ip (4 byte array)
                returns: none

        close
                Closes the socket.
                args:    none
                returns: none

        listen
                When TCP server use listen to establish connection
                args:    none
                returns: none

        connect
                When TCP client use connect to establish connection
                args:    none
                returns: none

        TX (dataptr, size)
                Transmits data
                args:    dataptr - byte array to be transmitted
                         size - size of the array
                returns: none

        RX (dataptr, size, block) : ret_size
                Receives data
                args:    dataptr - byte array received data is written to
                         size - the max size of the byte array
                         block - set to true to block (wait forever) for data
                returns: ret_size - the actual size of data received

        read_rsr : ret_val
                Read the receive size register. A non-zero indicates RX data.
                args:    none
                returns: ret_val - the value of the receive size register

        con_est : ret_val
                Returns true if a connection has been established
                args:    none
                returns: ret_val - true if connection has been established

        read(a0, a1) : ret_val
                Read the byte at address a0, a1
                args:    a0 - First byte of the address
                         a1 - Second byte of the address
                returns: ret_val - the byte from address a0,a1

        write(a0, a1, dout)
                Write a byte to address a0,a1
                args:    a0 - First byte of the address
                         a1 - Second byte of the address
                         dout - Data byte to write
                returns: none

        mode(opmode)
                Set the opmode of the W5100
                args:    opmode - New value of the opmode (valid from 0-7)
                returns: none              

API Instructions:
        TCP Server:
                object_name.start
                object_name.open(1, sport, dport, @dest_ip)
                object_name.listen
                repeat
                while object_name.con_est == false

                'Then call RX and TX function to read/write data

        TCP Client:
                object_name.start
                object_name.open(1, sport, dport, @dest_ip)
                object_name.connect
                repeat
                while object_name.con_est == false

                'Then call RX and TX function to read/write data
                
         UDP:
                object_name.start
                object_name.open(0, sport, dport, @dest_ip)

                'Then call RX and TX function to read/write data
                
                '******************************************************
                'NOTE: UDP packet reads include the UDP header
                '******************************************************

Performace measurements:
Size:
        Default of 2048 byte buffer for the socket. Thus, never TX more than 2K bytes.
        Never let the RX size be more than 2K bytes.

Speed performance (TCP w/ 256 byte payload):
        Write speed 7.5KB/sec (payload data)
        Read speed 7.8KB/sec (payload data)
        SPI write assembly takes 13.6us to write one byte
        SPI write assembly + spin functions take 58us (ideal)
        SPI read assembly take 14.1us to read one byte
        SPI read assembly + spin functions take 56us (ideal)

Storage:
        ASM portion is 396 bytes
        Entire driver is 326 longs with 8 longs variable

Detailed Change Log
--------------------
v1.1 (8.18.08)
- Removed read_sr function (only currently used to determine connection status)
- Added the conn_est function

v1.0 (5.14.08)
- Changed the pins to reflect the new PCB design.
- Driver can now select the W5100 mode.

v0.9 (4.22.08)
- Start pub function now has no args and takes the device out of
  reset.
- Added a new pub function called init which does include the
  ip, mac, gway, and subnet args.

v0.8 (4.1.08)
- Changed the start pub function to include the ip, mac, gway, and
  subnet args. 

v0.7 (3.18.08)
- Improved the RX function
- More detail in the comment header

v0.6 (3.16.08)
- Beta release

//////////////////////////////////////////////////////////////////////}}

'///////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////
CON

  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000 + 0000
   
  'Constants for UDP for TCP port
  UDP = 0
  TCP = 1
   
  'Constants used to make the code more readable
  SPI_RD = 1
  SPI_WR = 2
  SPI_DONE = 0

VAR

  long  SPIRW, add0, add1, dataout, datain
  long  socket_type
  long  cogon, cog

OBJ


PUB start : okay

  '' This is the public start function. It starts
  '' a new cog at the assembly entry point

  'Start the SPI code in a new COG
  stop
  okay := cogon := (cog := cognew(@entry, @SPIRW)) > 0

PUB stop

'' Stops driver - frees a cog

  if cogon~
    cogstop(cog)

PUB init (gway, subnet, ip, mac)

  'Init the registers
  'Gateway
  write($00,$01,byte[gway])
  write($00,$02,byte[gway+1])
  write($00,$03,byte[gway+2])
  write($00,$04,byte[gway+3])
  
  'Subnet
  write($00,$05,byte[subnet])
  write($00,$06,byte[subnet+1])
  write($00,$07,byte[subnet+2])
  write($00,$08,byte[subnet+3])

  'MAC
  write($00,$09,byte[mac])
  write($00,$0a,byte[mac+1])
  write($00,$0b,byte[mac+2])
  write($00,$0c,byte[mac+3])
  write($00,$0d,byte[mac+4])
  write($00,$0e,byte[mac+5])

  'IP
  write($00,$0f,byte[ip])
  write($00,$10,byte[ip+1])
  write($00,$11,byte[ip+2])
  write($00,$12,byte[ip+3])

PUB open (type, source_port, dest_port, dest_ip) | temp, temp2

  '' This function will open either a TCP or UDP socket based on
  '' the type arguement. The source_port and dest_port are the
  '' source and destination port numbers. The last arguement
  '' is meant to be a four byte array that stores the destination
  '' IP. We could have just made this a long arg but this
  '' format is more human friendly.
  
  'Set the socket type so that other function know if we are
  'TCP or UDP
  socket_type := type

  'Configure socket for UDP for TCP (TCP is the default)
  if(socket_type == UDP)
    write($04,$00,2)
  else
    write($04,$00,1)

  'Configure source port
  temp := (source_port >> 8) & $FF
  temp2 := source_port & $FF
  write($04,$04,temp)
  write($04,$05,temp2)

  'Configure dest port
  temp := (dest_port >> 8) & $FF
  temp2 := dest_port & $FF
  write($04,$10,temp)
  write($04,$11,temp2)

  'Dest IP
  write($04,$0c,byte[dest_ip])
  write($04,$0d,byte[dest_ip+1])
  write($04,$0e,byte[dest_ip+2])
  write($04,$0f,byte[dest_ip+3])
  
  'Open socket
  write($04,$01,1)

PUB close

  '' Close the socket.

  write($04,$01,$10)

PUB listen

  '' This function will initiate a TCP listen. Call this function
  '' for a TCP server to listen for a connection
  
  if(socket_type == TCP)
    write($04,$01,2)

PUB connect

  '' When TCP this function will establish a connection with a server.

  if(socket_type == TCP)
    write($04,$01,4) 

PUB TX (dataptr, size) | tptr, offset, startadd, a0, a1, counter, temp 

  '' Call this function to send data via the W5100

  'Read the offset so we know what the starting address is
  'of the TX buffer.
  tptr := read($04,$24)         'Read the transmit pointer
  tptr := tptr << 8
  tptr += read($04,$25)
  offset := tptr & $7FF         'Socket 0 has 2K of buffer and that determines the mask here of $7FF
  startadd := $4000 + offset    'Add the offset to the starting address of socket 0

  'Write the data to the W5100 internal memory buffer
  repeat counter from 0 to size-1
    a1 := startadd & $FF
    a0 := (startadd & $FF00) >> 8
    write(a0, a1, byte[dataptr])
    offset++
    offset := offset & $7FF     'Socket 0 has 2K of buffer and that determines the mask here of $7FF
    startadd := $4000 + offset  'Add the offset to the starting address of socket 0 
    dataptr++

  'Update the offset counter and write it back to the W5100
  tptr += counter
  temp := (tptr & $FF00) >> 8
  write($04,$24,temp)
  temp := tptr & $FF
  write($04,$25,temp)

  'Tell the W5100 to write the data
  write($04,$01,$20)

PUB RX (dataptr, size, block) : ret_size | offset, startadd, a0, a1, counter, rdptr, temp, tempsize

  '' Call this function to receive data via the W5100.
  '' The function will return the number of bytes read.
  '' Set block == true if you want the function to block waiting for data.
  ''
  '' This function will return the actual number of bytes read.

  'Block waiting for the W5100 to tell us that there is data.
  'Technically we are reading the number of bytes that have
  'been received
  if(block == true)
    repeat
      temp := read($04,$26)
      temp := temp << 8
      temp += read($04,$27)
    while temp == 0             'Wait as long as we have received zero bytes

  'If the user wants a non-block RX call then we will return immediatly
  'with a size of zero if there is no data to receive.
  else
    temp := read($04,$26)
    temp := temp << 8
    temp += read($04,$27)
    if(temp == 0)
      return 0

  'Compute the starting address
  rdptr := read($04,$28)        'Read the receive pointer
  rdptr := rdptr << 8
  rdptr += read($04,$29)
  offset := rdptr & $7FF        'Socket 0 has 2K of buffer and that determines the mask here of $7FF 
  startadd := $6000 + offset    'Add the offset to the starting address of socket 0

  'Determine how many bytes we need to read.
  tempsize := read_rsr
  if(tempsize > size)
    ret_size := size
  else
    ret_size := tempsize

  'Now we read the data from the W5100 and write it to the array
  'pointed to by dataptr.
  repeat counter from 0 to ret_size-1
    a1 := startadd & $FF
    a0 := (startadd & $FF00) >> 8 
    byte[dataptr] := read(a0,a1)
    dataptr++
    offset++
    offset := offset & $7FF     'Socket 0 has 2K of buffer and that determines the mask here of $7FF
    startadd := $6000 + offset  'Add the offset to the starting address of socket 0

  'Need to increment the rdptr by the number of bytes acutally read
  rdptr += ret_size
    
  'Then write the value of the new pointer back
  temp := (rdptr & $FF00) >> 8
  write($04,$28,temp)
  temp := rdptr & $FF
  write($04,$29,temp)

  'Tell the W5100 that we have read the data
  write($04,$01,$40)

  'Issue the read command to the W5100 which just updates registers.
  'We will block waiting for the W5100 to finish.
  'This takes very little time but we will check it just to be sure.
  repeat
    temp := read($04,$01)
  while temp <> $00

PUB read_rsr : ret_val

  '' Read the receive size register.
  '' This will return the value in the receive size register in the W5100.
  '' A non-zero value indicates that there is data to be read.

  ret_val := read($04,$26)
  ret_val := ret_val << 8
  ret_val += read($04,$27)

PUB con_est : ret_val

  '' Call this function to determine if a TCP connection has been
  '' established. The function will return true if a connection
  '' has been established.

  if(read($04,$03) == $17)
    ret_val := true
  else
    ret_val := false
  
PUB read(a0, a1) : ret_val

  '' Call this function to read a byte from the W5100 via SPI.
  '' The arguements are two bytes that contains address byte 0
  '' and address byte 1. See W5100 data sheet for what registers
  '' these actually address. The function will return the byte
  '' that came via SPI from the W5100.

  'Read data from the address specified in the arguements
  add0 := a0                    'Set the arguements to the global
  add1 := a1                    'variables values
  SPIRW := SPI_RD               'Set SPIRW to the read value

  'Wait until the driver clears SPIRW. This indicates that it is done.
  repeat until SPIRW == SPI_DONE

  'The driver wrote the result to datain. Thus, that is the value
  'that we will return
  ret_val := datain

PUB write(a0, a1, dout)

  '' Call this function to write a byte from the W5100 via SPI.
  '' The arguements are three bytes that contains address byte 0,
  '' address byte 1 and the data we wish to write. See W5100
  '' data sheet for what registers these actually address.
  '' Note that the datain global variable will be overwritten
  '' during this process. 

  'Write data to the address specified in the arguements
  add0 := a0                    'Set the arguements to the global
  add1 := a1                    'variables values 
  dataout := dout
  SPIRW := SPI_WR               'Set SPIRW to the write value

  'Wait until the driver clears SPIRW. This indicates that it is done.
  repeat until SPIRW == SPI_DONE

PUB mode(opmode) | temp

  '' Call this function with the value wanted for opmode before
  '' the start function is called. If the start function has
  '' already been called then this function will do nothing.

  if (cogon == 0)               'Be sure driver not already running
    DIRA |= $70000              'Set pins 16-18 to output
    temp := opmode << 16        'Shift our arg 16 bits left
    OUTA &= $FFF8FFFF           'Clear all outputs on pins 16-18
    OUTA |= opmode              'Set only required pins on pins 16-18
 
DAT
'*****************************************************
'* Assembly for reading/writing via SPI to the W5100 *
'*****************************************************
        org
        
entry   'Entry point for this driver. Start a new COG here.
        'Setup the IO and create masks.
        mov             t1,#0                   'Always make t1 equal to 0

        mov             ssm,#$10                'Create the mask for ss
        shl             ssm,#16

        mov             clkm,#$8                'Create the mask for clk
        shl             clkm,#16

        mov             doutm,#$20              'Create the mask for dout
        shl             doutm,#16

        mov             dinm,#$40               'Create the mask for din
        shl             dinm,#16

        mov             rstm,#$80               'Create the mask for reset
        shl             rstm,#16                        

        mov             t2,outa                 'Copy the outa register to t2
        or              t2,ssm                  'When ss goes output we want it high
        andn            t2,rstm                 'When reset goes output we want it low        
        andn            t2,clkm                 'When clk goes output we want it low
        andn            t2,doutm                'When dout goes output we want it low
        mov             outa,t2                 'Write the modified register back to outa        
        
        mov             t2,dira                 'Copy the dira register to t2
        or              t2,ssm                  'Set outputs
        or              t2,clkm
        or              t2,doutm
        or              t2,rstm
        mov             dira,t2                 'Write the modified value back to dira

        mov             t2,outa                 'Take the W5100 out of reset
        or              t2,rstm
        mov             outa,t2
        
        
        
SPILoop 'Main loop. Waits for SPIRW to be non-zero.
        'Reads arguements from main memory. Begins to set up MOSI pin.
        rdlong          t2,par                  'Grab the SPIRW value from main memory        
        tjz             t2,#SPILoop             'If this value is zero then we are done

        mov             tpar,par                'Get a temporary copy of the parameter pointer
        add             tpar,#4                 'Add one to the pointer to get us to point to add0
        rdlong          ta0,tpar                'Get address0 from main memory
        add             tpar,#4                 'Add one to the pointer to get us to point to add1
        rdlong          ta1,tpar                'Get address1 from main memory
        add             tpar,#4                 'Add one to the pointer to get us to point to dataout
        rdlong          tdout,tpar              'Get dataout from main memory

        djnz            t2,#sbranch7            'Check to see if we should read or write
        andn            outa,doutm              'If we read then set dout to $0F
        jmp             #firstnibble
        
sbranch7 
        or              outa,doutm              'If we write then set dout for $F0

firstnibble
        andn            outa,ssm                'Drop the SS bit
        mov             t2,#4
        
sloop1  'Here we will send the first byte. It is $0F for a read and $F0 for a write.                                        
        or              outa,clkm               'Set clock high
        andn            outa,clkm               'Clear clock low
        djnz            t2,#sloop1              'Repeat this four times (first nibble is $0 or $F)

        rdlong          t2,par                  'Grab the SPIRW value from main memory
        djnz            t2,#sbranch8            'Determine what the second nibble should be (either $0 or $F)
        or              outa,doutm              'Set the MOSI if we are a read operation
        jmp             #secnibble

sbranch8
        andn            outa,doutm              'Clear the MOSI if we are a write operation

secnibble
        mov             t2,#4                   'Reset the loop counter
        
sloop2
        or              outa,clkm               'Set clock high
        andn            outa,clkm               'Clear clock low
        djnz            t2,#sloop2              'Repeat this four times (fsecond nibble is $0 or $F)

        'Send the second byte which is the first byte of the address (address 0)        
        mov             t2,#8                   'Init the loop counter               
sloop3
        mov             t3,#$80                 'We will look at the MSb every time
        and             t3,ta0                  'AND the address0 value with $80 to determine if the MSb is set
        tjz             t3,#sbranch1            'Test to see if the result is zero
        or              outa,doutm              'If not then we set the MOSI bit
        jmp             #sbranch2
sbranch1
        andn            outa,doutm              'If it is zero then we clear the MOSI bit
sbranch2
        shl             ta0,#1                  'Shift the register left one so that the next bit is the new MSb.
        or              outa,clkm               'Set the clock
        andn            outa,clkm               'Clear the clock
        djnz            t2,#sloop3              'Do this eight times

        'Send the third byte which is the second byte of the address (address 1)
        mov             t2,#8                   'Init the loop counter             
sloop4
        mov             t3,#$80                 'We will look at the MSb every time
        and             t3,ta1                  'AND the address0 value with $80 to determine if the MSb is set
        tjz             t3,#sbranch3            'Test to see if the result is zero
        or              outa,doutm              'If not then we set the MOSI bit
        jmp             #sbranch4
sbranch3
        andn            outa,doutm              'If it is zero then we clear the MOSI bit
sbranch4
        shl             ta1,#1                  'Shift the register left one so that the next bit is the new MSb.
        or              outa,clkm               'Set the clock
        andn            outa,clkm               'Clear the clock
        djnz            t2,#sloop4              'Do this eight times

        'The fourth byte is special. To fit this driver in under 512 bytes the read
        'and the write operations both happen here. So, even during a read operation
        'we will write the value in the dataout global variable. This is ok because
        'the W5100 will happily ignore it (as it knows we are reading). Also, when
        'performing a write operation the datain global variable will be overwritten.
        'The W5100 data sheet claims that in this case the value will always be $03.
        mov             t2,#8                   'Init the loop counter                   
        mov             tdin,#0                 'Clear the tdin variable
sloop5                                          'Here we write whatever value is in dataout
        mov             t3,#$80                 'We will look at the MSb every time
        and             t3,tdout                'AND the address0 value with $80 to determine if the MSb is set 
        tjz             t3,#sbranch5            'Test to see if the result is zero
        or              outa,doutm              'If not then we set the MOSI bit
        jmp             #sbranch6
sbranch5                                         
        andn            outa,doutm              'If it is zero then we clear the MOSI bit
sbranch6
        shl             tdout,#1                'Shift the register left one so that the next bit is the new MSb.                                  
        mov             t3,dinm                 'Start shifting in data from MISO. Move the mask for P16 to register t3
        or              outa,clkm               'Set the clock               
        shl             tdin,#1                 'First shift tdin left. Now we can set/clear the new LSb.
        and             t3,ina                  'Determine if P16 is set/clear
        tjz             t3,#sbranch9
        or              tdin,#1                 'If it is set then set the LSb of tdin. If clear we do nothing
sbranch9                                        'because shifting the value left will produce a clear bit
        andn            outa,clkm               'Clear the clock
        djnz            t2,#sloop5              'Repeat this eight times

        or              outa,ssm                'Set SS. This ends the data transaction
        andn            outa,doutm              'Clear MOSI pin.

        mov             tpar,par                'Move the tdin value to the global datain memory location
        add             tpar,#16
        wrlong          tdin,tpar
        
        wrlong          t1,par                  'Clear the SPIRW variable to indicate we are done        
        jmp             #SPILoop                'Jump back to the main loop
        
        'To here is 396 bytes (includes the reserves at the bottom)
        
tpar    res   1         'temp variable for the par global memory pointer
t1      res   1         'temp variable 1 which is always zero         
t2      res   1         'temp variable 2
t3      res   1         'temp variable 3 
ta0     res   1         'temp address0
ta1     res   1         'temp address1
tdin    res   1         'temp datain
tdout   res   1         'temp dataout
ssm     res   1         'SS pin mask
clkm    res   1         'CLK pin mask
doutm   res   1         'MOSI pin mask
dinm    res   1         'MISO pin mask
rstm    res   1         'RST pin mask